
;;; From cmulisp. Most probably 19f.

(defpackage :cmu)

(in-package :cmu)

;;;; Float printing.

;;;
;;;  Written by Bill Maddox
;;;
;;;
;;;
;;; FLONUM-TO-STRING (and its subsidiary function FLOAT-STRING) does most of 
;;; the work for all printing of floating point numbers in the printer and in
;;; FORMAT.  It converts a floating point number to a string in a free or 
;;; fixed format with no exponent.  The interpretation of the arguments is as 
;;; follows:
;;;
;;;     X        - The floating point number to convert, which must not be
;;;                negative.
;;;     WIDTH    - The preferred field width, used to determine the number
;;;                of fraction digits to produce if the FDIGITS parameter
;;;                is unspecified or NIL.  If the non-fraction digits and the
;;;                decimal point alone exceed this width, no fraction digits
;;;                will be produced unless a non-NIL value of FDIGITS has been
;;;                specified.  Field overflow is not considerd an error at this
;;;                level.
;;;     FDIGITS  - The number of fractional digits to produce. Insignificant
;;;                trailing zeroes may be introduced as needed.  May be
;;;                unspecified or NIL, in which case as many digits as possible
;;;                are generated, subject to the constraint that there are no
;;;                trailing zeroes.
;;;     SCALE    - If this parameter is specified or non-NIL, then the number
;;;                printed is (* x (expt 10 scale)).  This scaling is exact,
;;;                and cannot lose precision.
;;;     FMIN     - This parameter, if specified or non-NIL, is the minimum
;;;                number of fraction digits which will be produced, regardless
;;;                of the value of WIDTH or FDIGITS.  This feature is used by
;;;                the ~E format directive to prevent complete loss of
;;;                significance in the printed value due to a bogus choice of
;;;                scale factor.
;;;     ALLOW-OVERFLOW-P -
;;;                This parameter, defaulting to non-NIL, allows us to
;;;                overflow any width constraint if necessary to
;;;                produce a reasonable output.  If NIL, then we will
;;;                not exceed the width constraint, even if that means
;;;                we only print out leading zeroes.  Generally
;;;                intended to be set by FIXED-FORMAT-AUX so we don't
;;;                exceed the specified width when printing small
;;;                numbers.
;;;
;;; Returns:
;;; (VALUES DIGIT-STRING DIGIT-LENGTH LEADING-POINT TRAILING-POINT DECPNT)
;;; where the results have the following interpretation:
;;;
;;;     DIGIT-STRING    - The decimal representation of X, with decimal point.
;;;     DIGIT-LENGTH    - The length of the string DIGIT-STRING.
;;;     LEADING-POINT   - True if the first character of DIGIT-STRING is the
;;;                       decimal point.
;;;     TRAILING-POINT  - True if the last character of DIGIT-STRING is the
;;;                       decimal point.
;;;     POINT-POS       - The position of the digit preceding the decimal
;;;                       point.  Zero indicates point before first digit.
;;;     ROUNDOFF        - True If, due to rounding, the exponent for the
;;;                       printed result differs from the real
;;;                       exponent of the number.  For example, .9999
;;;                       has an exponent of -1, but the printed
;;;                       result could be 1.000 with an exponent of 0.
;;;
;;; NOTE:  FLONUM-TO-STRING goes to a lot of trouble to guarantee accuracy.
;;; Specifically, the decimal number printed is the closest possible 
;;; approximation to the true value of the binary number to be printed from 
;;; among all decimal representations  with the same number of digits.  In
;;; free-format output, i.e. with the number of digits unconstrained, it is 
;;; guaranteed that all the information is preserved, so that a properly-
;;; rounding reader can reconstruct the original binary number, bit-for-bit, 
;;; from its printed decimal representation. Furthermore, only as many digits
;;; as necessary to satisfy this condition will be printed.
;;;
;;;
;;; FLONUM-TO-DIGITS actually generates the digits for positive
;;; numbers.  See below for comments.

(defvar *digits* "0123456789")

(defun flonum-to-string (x &key width fdigits scale fmin (num-expt 0 num-expt-p)
			        (allow-overflow-p t))
  (setf x (abs x))
  (cond ((zerop x)
	 ;;zero is a special case which float-string cannot handle
	 (if fdigits
	     (let ((s (make-string (1+ fdigits) :initial-element #\0)))
	       (setf (schar s 0) #\.)
	       (values s (length s) t (zerop fdigits) 0))
	     (values "." 1 t t 0)))
	(t
	 (flet ((fixup-flonum-to-digits (x &optional (expt 0) position relativep)
		  (multiple-value-bind (e s)
		      (flonum-to-digits x position relativep)
		    (values (- e expt) s e))))
	   (multiple-value-bind (e string printed-e)
	       (if fdigits
		   (fixup-flonum-to-digits x
					   num-expt
					   (+ (min (- (+ fdigits (or scale 0)))
						   (- (or fmin 0)))
					      num-expt))
		   (if (and width (> width 1))
		       (let ((ndigits (1- width)))
			 #+(or)
			 (progn
			   (format t "ndigits = ~A~%" ndigits)
			   (format t "scale   = ~A~%" scale))
			 (when (and scale (< scale 0))
			   (setf ndigits (+ ndigits scale)))
			 (when (minusp num-expt)
			   (incf ndigits (- num-expt)))
			 (fixup-flonum-to-digits x num-expt
						 ndigits
						 t))
		       (fixup-flonum-to-digits x num-expt)))
	     (let ((stream (make-string-output-stream))
		   (printed-roundoff-p (and num-expt-p (/= num-expt printed-e))))
	       (when (and num-expt-p (/= num-expt printed-e))
		 ;; The actual exponent of the number differs from the
		 ;; printed exponent.  This happens for something like
		 ;; (format nil "~11,3,2,0,'*,,'EE" .9999).  With only
		 ;; 3 fraction digits, .9999 gets rounded to 1.000 But
		 ;; the exponent for .9999 is -1, and the exponent for
		 ;; the printed result is actually 1.  We need to
		 ;; decrement e by 1 to account for this.
		 (decf e))
	       #+(or)
	       (progn
		 (format t "e = ~S~%" e)
		 (format t "roundoff = ~S~%" printed-roundoff-p))
	       
	       (incf e (or scale 0))
	       (if (plusp e)
		   (progn
		     (write-string string stream :end (min (length string)
							   e))
		     (dotimes (i (- e (length string)))
		       (write-char #\0 stream))
		     (write-char #\. stream)
		     (write-string string stream :start (min (length string)
							     e))
		     (when fdigits
		       (dotimes (i (- fdigits
				      (- (length string) 
					 (min (length string) e))))
			 (write-char #\0 stream))))
		   (progn
		     (write-string "." stream)
		     ;; Write out the leading zeroes.  If fmin is set,
		     ;; we need all of them.  But if fdigits is given
		     ;; and is smaller than -e, we only want fdigits
		     ;; to be output.  That way we don't print too
		     ;; many leading zeroes if the number is too
		     ;; small.
		     ;;
		     ;; Also print them all out if :allow-overflow-p
		     ;; is set or there's no width constraint.
		     #+(or)
		     (progn
		       (format t "fmin = ~S~%" fmin)
		       (format t "fdigits = ~S~%" fdigits)
		       (format t "width = ~S~%" width)
		       (when width
			 (format t "min = ~S~%" (min (- e) width))))

		     (let ((leading-zeros
			    (if (or fmin (null fdigits))
				(if (or allow-overflow-p (null width))
				    (- e)
				    (min (- e) (1- width)))
				(min (- e) fdigits))))
		       (dotimes (i leading-zeros)
			 (write-char #\0 stream))
		       ;; If we're out of room (because fdigits is too
		       ;; small), don't print out our string.  This
		       ;; fixes things like (format nil "~,2f" 0.001).
		       ;; We should print ".00", not ".001".  But if
		       ;; fmin is set, we want to print out something.
		       (when (or (null fdigits)
				 (plusp (+ e fdigits))
				 fmin)
			 ;; But only print the whole string if there's
			 ;; no width constraint or if we're allowed to
			 ;; exceed the width.
			 (if (or allow-overflow-p (null width))
			     (write-string string stream)
			     (write-string string stream
					   :end (min (- width leading-zeros 1)
						     (length string)))))
		       (when fdigits
			 (dotimes (i (+ fdigits e (- (length string))))
			   (write-char #\0 stream))))))
	       (let ((string (get-output-stream-string stream)))
		 (values string (length string)
			 (char= (char string 0) #\.)
			 (char= (char string (1- (length string))) #\.)
			 (position #\. string)
			 printed-roundoff-p))))))))


;;;; Entry point for the float printer.

;;; Entry point for the float printer as called by PRINT, PRIN1, PRINC,
;;; etc.  The argument is printed free-format, in either exponential or 
;;; non-exponential notation, depending on its magnitude.
;;;
;;; NOTE: When a number is to be printed in exponential format, it is scaled in
;;; floating point.  Since precision may be lost in this process, the
;;; guaranteed accuracy properties of FLONUM-TO-STRING are lost.  The
;;; difficulty is that FLONUM-TO-STRING performs extensive computations with
;;; integers of similar magnitude to that of the number being printed.  For
;;; large exponents, the bignums really get out of hand.  If bignum arithmetic
;;; becomes reasonably fast and the exponent range is not too large, then it
;;; might become attractive to handle exponential notation with the same
;;; accuracy as non-exponential notation, using the method described in the
;;; Steele and White paper.


;;; PRINT-FLOAT-EXPONENT  --  Internal
;;;
;;;    Print the appropriate exponent marker for X and the specified exponent.
;;;
(defun print-float-exponent (x exp stream)
  (declare (float x) (integer exp) (stream stream))
  (let ((*print-radix* nil))
    ;; CLHS 22.3.3.2 for ~E says (near the bottom) that
    ;;
    ;;    If all of w, d, and e are omitted, then the effect is to
    ;;    print the value using ordinary free-format
    ;;    exponential-notation output; prin1 uses a similar format for
    ;;    any non-zero number whose magnitude is less than 10^-3 or
    ;;    greater than or equal to 10^7. The only difference is that
    ;;    the ~E directive always prints a plus or minus sign in front
    ;;    of the exponent, while prin1 omits the plus sign if the
    ;;    exponent is non-negative.
    ;;
    ;; So we don't want a + sign for the exponent.
    (if (typep x *read-default-float-format*)
	(unless (eql exp 0)
	  (format stream "e~D" exp))
	(format stream "~C~D" 
		(etypecase x
		  (single-float #\f)
		  (double-float #\d)
		  (short-float #\s)
		  (long-float #\l)
		  #+double-double
		  (double-double-float #\w))
		exp))))


;;; FLOAT-FORMAT-NAME  --  Internal
;;;
;;;    Return the string name of X's float format.
;;;
(defun float-format-name (x)
  (declare (float x))
  (etypecase x
    (single-float "SINGLE-FLOAT")
    (double-float "DOUBLE-FLOAT")
    (short-float "SHORT-FLOAT")
    (long-float "LONG-FLOAT")
    #+double-double
    (double-double-float "DOUBLE-DOUBLE-FLOAT")))


;;; OUTPUT-FLOAT-INFINITY  --  Internal
;;;
;;;    Write out an infinity using #. notation, or flame out if
;;; *print-readably* is true and *read-eval* is false.
;;;
(defun output-float-infinity (x stream)
  (declare (float x) (stream stream))
  (cond (*read-eval*
	 (write-string "#." stream))
	(*print-readably*
	 (error 'print-not-readable :object x))
	(t
	 (write-string "#<" stream)))
  (write-string "EXT:" stream)
  (write-string (float-format-name x) stream)
  (write-string (if (plusp x) "-POSITIVE-" "-NEGATIVE-")
		stream)
  (write-string "INFINITY" stream)
  (unless *read-eval*
    (write-string ">" stream)))


;;; OUTPUT-FLOAT-NAN  --  Internal
;;;
;;;    Output a #< NaN or die trying.
;;;
(defun output-float-nan (x stream)
  (print-unreadable-object (x stream)
    (write-string (float-format-name x) stream)
    (write-string (if (float-trapping-nan-p x) " Trapping" " Quiet") stream)
    (write-string " NaN" stream)))


(defconstant output-float-free-format-exponent-min -3
  "Minimum power of 10 that allows the float printer to use free format,
   instead of exponential format.  See section 22.1.3.1.3: Printing Floats
   in the ANSI CL standard.")
(defconstant output-float-free-format-exponent-max 8
  "Maximum power of 10 that allows the float printer to use free format,
   instead of exponential format.  See section 22.1.3.1.3: Printing Floats
   in the ANSI CL standard.")

;;; OUTPUT-FLOAT  --  Internal
;;;
;;;    Functioned called by OUTPUT-OBJECT to handle floats.
;;;

(defun output-float (x stream)
  (cond
   ((si::float-infinity-p x)
    (si::output-float-infinity x stream))
   ((si::float-nan-p x)
    (si::output-float-nan x stream))
   #+(and nil double-double)
   ((typep x 'double-double-float)
    ;; This is a temporary hack until we get more double-double float
    ;; stuff in place so that we don't need this.  And being able to
    ;; print double-double's now is quite handy while we debug the
    ;; rest of the code.
    (output-double-double-float x stream))
   (t
    (let ((x (cond ((minusp (float-sign x))
		    (write-char #\- stream)
		    (- x))
		   (t
		    x))))
      (cond
       ((zerop x)
	(write-string "0.0" stream)
	(print-float-exponent x 0 stream))
       (t
	(output-float-aux x stream
			  output-float-free-format-exponent-min
			  output-float-free-format-exponent-max)))))))

;; Smallest possible (unbiased) exponents
(defconstant double-float-min-e
  (nth-value 1 (decode-float least-positive-double-float)))
(defconstant single-float-min-e
  (nth-value 1 (decode-float least-positive-single-float)))
(defconstant long-float-min-e
  (nth-value 1 (decode-float least-positive-long-float)))
#+double-double
(defconstant double-double-float-min-e double-float-min-e)

;; Exact powers of ten.  Must be large enough to cover the range from
;; least-positive-double-float to most-positive-double-float
;;(declaim (type (simple-array integer (326)) *powers-of-ten*))
(defparameter *powers-of-ten*
  (make-array 5000 ;;326
	      :initial-contents
	      (let (p)
		(dotimes (k 5000 #|326|# (nreverse p))
		  (push (expt 10 k) p)))))

;;; Implementation of Figure 1 from "Printing Floating-Point Numbers
;;; Quickly and Accurately", by Burger and Dybvig, 1996.  As the
;;; implementation of the Dragon algorithm above says: "DO NOT EVEN
;;; THINK OF ATTEMPTING TO UNDERSTAND THIS CODE WITHOUT READING THE
;;; PAPER!"  This is a more idiomatic Lisp translation by CSR of the
;;; scheme implementation in that paper.
;;;
(defun flonum-to-digits (v &optional position relativep)
  ;; V is the number to be printed.  If RELATIVEP is NIL, POSITION is
  ;; the number of digits to the left of the decimal point where we
  ;; want to stop printing.  If RELATIVEP is non-NIL, POSITION is the
  ;; total number of digits we want printed.
  ;;
  ;; Two values are returned: k, and the digit string, without a
  ;; decimal point.  k is the index into the string, before which the
  ;; decimal point would go.
  (let ((print-base 10)			; B
	(float-radix 2)			; b
	(float-digits (float-digits v)) ; p
	(min-e
	 (etypecase v
	   (single-float single-float-min-e)
	   (double-float double-float-min-e)
	   (long-float long-float-min-e)
	   #+double-double
	   (double-double-float double-double-float-min-e))))
    (multiple-value-bind (f e)
	(integer-decode-float v)
      (let ( ;; FIXME: these even tests assume normal IEEE rounding
	    ;; mode.  I wonder if we should cater for non-normal?
	    (high-ok (evenp f))
	    (low-ok (evenp f))
	    (result (make-array 50 :element-type 'base-char
				:fill-pointer 0 :adjustable t)))
	(labels ((flog (x)
		   (declare (type (float (0.0)) x))
		   (let ((xd (etypecase x
			       (single-float
				(float x 1d0))
			       (double-float
				x)
			       (long-float
				x)
			       #+double-double
			       (double-double-float
				(double-double-hi x)))))
		     (ceiling (- (the (double-float -400d0 400d0) (log xd 10d0))
				 1d-10))))
		 (fixup (r s m+ m- k)
		   (if (if high-ok
			   (>= (+ r m+) s)
			   (> (+ r m+) s))
		       (values (+ k 1) (generate r (* s print-base) m+ m-))
		       (values k (generate r s m+ m-))))
		 (scale (r s m+ m-)
		   (let* ((est (flog v))
			  (scale (the integer (aref *powers-of-ten* (abs est)))))
		     (if (>= est 0)
			 (fixup r (* s scale) m+ m- est)
			 (fixup (* r scale) s (* m+ scale) (* m- scale) est))))
		 (generate (r s m+ m-)
		   (multiple-value-bind (d r)
		       (truncate (* r print-base) s)
		     (let ((m+ (* m+ print-base))
			   (m- (* m- print-base)))
		       (let ((tc1 (or (< r m-) (and low-ok (= r m-))))
			     (tc2 (let ((test (+ r m+)))
				    (or (> test s)
					(and high-ok (= test s))))))
			 (cond
			   ((and (not tc1) (not tc2))
			    (vector-push-extend (char *digits* d) result)
			    ;; FIXME sucky tail recursion.  This whole
			    ;; kaboodle should be DO*/LOOPified.
			    (generate r s m+ m-))
			   ;; pedantically keeping all the conditions
			   ;; in so that we can move them around.
			   ((and (not tc1) tc2)
			    (vector-push-extend (char *digits* (1+ d)) result)
			    result)
			   ((and tc1 (not tc2))
			    (vector-push-extend (char *digits* d) result)
			    result)
			   ((and tc1 tc2)
			    (vector-push-extend (char *digits*
						      (if (< (* r 2) s) d (1+ d)))
						result)
			    result)))))))
	  (let (r s m+ m-)
	    (if (>= e 0)
		(let* ((be (expt float-radix e))
		       (be1 (* be float-radix)))
		  (if (/= f (expt float-radix (1-
					       float-digits)))
		      (setf r (* f be 2)
			    s 2
			    m+ be
			    m- be)
		      (setf r (* f be1 2)
			    s (* float-radix 2)
			    m+ be1
			    m- be)))
		(if (or (= e min-e) 
			(/= f (expt float-radix (1-
						 float-digits))))
		    (setf r (* f 2)
			  s (* (expt float-radix (- e)) 2)
			  m+ 1
			  m- 1)
		    (setf r (* f float-radix 2)
			  s (* (expt float-radix (- 1 e)) 2)
			  m+ float-radix
			  m- 1)))
	    (when position
	      (flet ((expt-ten (e)
		       (if (minusp e)
			   (/ (aref *powers-of-ten* (- e)))
			   (aref *powers-of-ten* e))))
		(when relativep
		  (let ((r+m (+ r m+)))
		    ;;(format t "r, s = ~A, ~A~%" r s)
		    (do ((k 0 (1+ k))
			 ;; running out of letters here
			 (l 1 (* l print-base)))
			((>= (* s l) r+m)
			 ;; k is now \hat{k}
			 ;;(format t "r>s, k = ~A~%" k)
			 (if (< (+ r (* s (/ (expt-ten (- k position))
					     2)))
				(* s (expt-ten k)))
			     (setf position (- k position))
			     (setf position (- k position 1)))))))
		(let ((low (max m- (/ (* s (expt-ten position)) 2)))
		      (high (max m+ (/ (* s (expt-ten position)) 2))))
		  (when (<= m- low)
		    (setf m- low)
		    (setf low-ok t))
		  (when (<= m+ high)
		    (setf m+ high)
		    (setf high-ok t)))))
	    (scale r s m+ m-)))))))


(defun output-float-aux (x stream e-min e-max)
  (multiple-value-bind (e string)
      (flonum-to-digits x)
    (cond
      ((< e-min e e-max)
       ;; free format
       (cond ((plusp e)
	      (write-string string stream :end (min (length string) e))
	      (dotimes (i (- e (length string)))
		(write-char #\0 stream))
	      (write-char #\. stream)
	      (write-string string stream :start (min (length string) e))
	      (when (<= (length string) e)
		(write-char #\0 stream))
	      (print-float-exponent x 0 stream))
	     (t
	      (write-string "0." stream)
	      (dotimes (i (- e))
		(write-char #\0 stream))
	      (write-string string stream)
	      (print-float-exponent x 0 stream))))
      (t
       ;; Exponential format
       (write-string string stream :end 1)
       (write-char #\. stream)
       (write-string string stream :start 1)
       ;; CLHS 22.1.3.1.3 says at least one digit must be printed
       ;; after the decimal point.
       (when (= (length string) 1)
	 (write-char #\0 stream))
       (print-float-exponent x (1- e) stream)))))


(defun accurate-scale-exponent (v)
  (declare (type float v))
  (if (zerop v)
      1
      (let ((float-radix 2)		; b
	    (float-digits (float-digits v)) ; p
	    (min-e
	     (etypecase v
	       (single-float single-float-min-e)
	       (double-float double-float-min-e)
	       (long-float long-float-min-e)
	       #+double-double
	       (double-double-float lisp::double-double-float-min-e))))
	(multiple-value-bind (f e)
	    (integer-decode-float v)
	  (let ( ;; FIXME: these even tests assume normal IEEE rounding
		;; mode.  I wonder if we should cater for non-normal?
		(high-ok (evenp f)))
	    ;; This scale function is basically the same as the one in
	    ;; FLONUM-TO-DIGITS, except we don't return the computed
	    ;; digits.  We only want the exponent.
	    (labels ((flog (x)
		       (declare (type (float (0.0)) x))
		       (let ((xd (etypecase x
				   (single-float
				    (float x 1d0))
				   (double-float
				    x)
				   (long-float
				    x)
				   #+double-double
				   (double-double-float
				    (double-double-hi x)))))
			 (ceiling (- (the (double-float -400d0 400d0) (log xd 10d0))
				     1d-10))))
		     (fixup (r s m+ k)
		       (if (if high-ok
			       (>= (+ r m+) s)
			       (> (+ r m+) s))
			   (+ k 1)
			   k))
		     (scale (r s m+)
		       (let* ((est (flog v))
			      (scale (the integer (aref *powers-of-ten* (abs est)))))
			 (if (>= est 0)
			     (fixup r (* s scale) m+ est)
			     (fixup (* r scale) s (* m+ scale) est)))))
	      (let (r s m+)
		(if (>= e 0)
		    (let* ((be (expt float-radix e))
			   (be1 (* be float-radix)))
		      (if (/= f (expt float-radix (1- float-digits)))
			  (setf r (* f be 2)
				s 2
				m+ be)
			  (setf r (* f be1 2)
				s (* float-radix 2)
				m+ be1)))
		    (if (or (= e min-e) 
			    (/= f (expt float-radix (1- float-digits))))
			(setf r (* f 2)
			      s (* (expt float-radix (- e)) 2)
			      m+ 1)
			(setf r (* f float-radix 2)
			      s (* (expt float-radix (- 1 e)) 2)
			      m+ float-radix)))
		(scale r s m+))))))))


;;;;;
;;;

(defun float-string (x) (flonum-to-string x :scale 1 :num-expt (accurate-scale-exponent x)))