
#define _GNU_SOURCE

#include <signal.h>
#include <semaphore.h>
#include <errno.h>
#include <unistd.h>
#include <fenv.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <setjmp.h>

volatile struct timespec tic_tac;

#define SIG_THR_RESTART SIGXCPU
#define SIG_SUSPEND SIGPWR

#define MAX_THR 6

pthread_t all_threads[MAX_THR];

void ABORT(char * msg)
{
  printf("\nERROR: %s\n", msg);
  fflush(NULL);
  exit(3);
}

void * busy(void * arg)
{
  int id = (long) arg;

  for (;;)
    {
      long doze_time;
      struct timespec now;

      clock_gettime(CLOCK_REALTIME, &now);

      tic_tac = now;
      pthread_yield();

      doze_time = (long)((200 * 1000) * ((double) random() / RAND_MAX));
      usleep(doze_time);
#if 1
      printf(" %d ", id);
      fflush(NULL);
#endif
    }
}

static sem_t GC_suspend_ack_sem;

static sigset_t suspend_handler_mask;

static void GC_suspend_handler(int sig)
{
  sem_post(&GC_suspend_ack_sem);

  printf("\\"); fflush(NULL);

  sigsuspend (&suspend_handler_mask);
}

static void GC_restart_handler(int sig)
{
  if (sig != SIG_THR_RESTART) ABORT("Bad signal in suspend_handler");

  printf("/"); fflush(NULL);
}

void stop_world(void)
{
  int ack_count = 0;
  int i;

  for (i = 0; i < MAX_THR; i++)
    {
      int result = pthread_kill(all_threads[i], SIG_SUSPEND);
      switch(result) {
      case ESRCH:
	printf("thread %d got lost!", i);
	break;
      case 0:
	break;
      default:
	ABORT("pthread_kill failed");
      }
    }

#if 0
  while (ack_count < MAX_THR)
    {
      int sem_count;

      sem_getvalue(&GC_suspend_ack_sem, &sem_count);
      ack_count += sem_count;
      usleep(3000);
    }
#else
  for (i = 0; i < MAX_THR; i++)
    {
    retry:
      if (0 != sem_wait(&GC_suspend_ack_sem)) {
	/* On Linux, sem_wait is documented to always return zero.*/
	/* But the documentation appears to be incorrect.         */
	if (errno == EINTR) {
	  /* Seems to happen with some versions of gdb.   */
	  goto retry;
	}
	ABORT("sem_wait for suspend handler failed");
      }
    }
#endif

  fflush(NULL);
  printf("\n-\n");
  fflush(NULL);
}

void start_world(void)
{
  int i;

  for (i = 0; i < MAX_THR; i++)
    {
      int result = pthread_kill(all_threads[i], SIG_THR_RESTART);
      switch(result) {
      case ESRCH:
	printf("thread %d got lost!", i);
	break;
      case 0:
	break;
      default:
	ABORT("pthread_kill failed");
      }
    }
  printf("\n+\n");
  fflush(NULL);
}

void setup(void)
{
  struct sigaction act;

  act.sa_flags = SA_RESTART;

  if (sem_init(&GC_suspend_ack_sem, 0, 0)) ABORT("sem_init failed");

  if (sigfillset(&act.sa_mask)) ABORT("sigfillset() failed");

  act.sa_handler = GC_suspend_handler;
  
  if (sigaction(SIG_SUSPEND, &act, NULL)) ABORT("Cannot set SIG_SUSPEND handler");
  
  act.sa_handler = GC_restart_handler;
  
  if (sigaction(SIG_THR_RESTART, &act, NULL)) {
    ABORT("Cannot set SIG_THR_RESTART handler");
  }
  
  if (sigfillset(&suspend_handler_mask)) ABORT("sigfillset() failed");
  
  if (sigdelset(&suspend_handler_mask, SIG_THR_RESTART)) ABORT("sigdelset() failed");
}

pthread_t fake_gc_thread;

void * fake_gc_monitor(void * arg)
{
  for (;;)
    {
      sleep(3);

      stop_world();
      usleep(10 * 1000); /* here goes the fake gc phase. */
      start_world();
    }
}

int main(int argc, char * argv[])
{
  struct timespec old_time = { 0, 0 };
  int except_mask =  FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW | FE_UNDERFLOW /* | FE_INEXACT  */;
  int status = feenableexcept(except_mask);
  int i = 0;
  long j = 0;
  int max = 20;
  jmp_buf jmp_env;
  
  if (argc > 1)
    {
      sscanf(argv[1], "%d", &max);
    }

  setup();

  all_threads[0] = pthread_self();

  for (j = 1; j < MAX_THR; j++)
    { 
      if (status = pthread_create(&all_threads[j], NULL, busy, (void *)j))
	{
	  printf("\npthread_create failed with status = %d\n", status);
	  return 1;
	}
    }

#if 1
  if (status = pthread_create(&fake_gc_thread, NULL, fake_gc_monitor, 0))
    {
      printf("\npthread_create failed with status = %d\n", status);
      return 1;
    }
#endif
  
  if ( _setjmp(jmp_env) )
    printf("\nsetjmp returned a non-zero value\n.");
  else
    for (i = 0; i < max; i++)
      {
	int mask = fegetexcept();

	if ( mask != except_mask )
	  {
	    printf("\nfegetexcept returned: %d\n", mask);
	    return 1;
	  }
	pthread_yield();
	{
	  long doze;

	  doze = 500 * 1000 + (long)((200 * 1000) * ((double) rand() / RAND_MAX));

	  usleep(doze);
	}
	if ((tic_tac.tv_sec <= old_time.tv_sec) && (tic_tac.tv_nsec <= old_time.tv_nsec))
	  {
	    printf("\ntimer stopped moving!\n");
	    return 2;
	  }
	else
	  old_time = tic_tac;
	printf(".\n");
	fflush(NULL);
      }
  printf("\nAt end fegetexcept() returned %d.\n", fegetexcept());
  return 0;
}
