;;; PUZZLE -- Forest Baskett's Puzzle benchmark, originally written in
;; Pascal. Solves a combinatorial puzzle by exploring the search
;; space, using arrays to represent state.
;;
;; Comments from Henry Baker: Puzzle solves a 3-dimensional packing
;; problem by attempting to pack pieces of 4 different types into a
;; 5x5x5 cube. The class of such packing problems is closely related
;; to the "bin-packing" and "knapsack" problems of complexity theory,
;; which are known to be NP-complete.
;;
;; The standard Gabriel code for Puzzle solves the problem by
;; "pre-rotating" all of the different puzzle pieces, so that these
;; rotations do not have to be performed during the actual
;; combinatorial search. Thus, each piece "class" has a number of
;; different piece "types" which result from different rotations of
;; the pieces. For example, the 4x2x1 piece has 6 distinct rotations,
;; while the 2x2x2 piece has only 1. The 5x5x5 puzzle cube itself is
;; represented as a section of an 8x8x8 cube of Boolean values, while
;; the various piece types (rotations) are represented as a vector of
;; Boolean values which is in correspondence with the representation
;; of the puzzle itself. By embedding a 5x5x5 cube within an 8x8x8
;; cube, a "border" is created which makes sure that the pieces stay
;; within the 5x5x5 boundaries.
;;
;; After initialization, the standard Puzzle code linearly searches
;; the puzzle representation for the smallest-numbered empty location.
;; It then tries all of the remaining pieces to see if they can be fit
;; into the puzzle in such a way that this empty location will be
;; filled. If a piece can be fitted, then the code performs a
;; depth-first search for the next empty location and the next piece
;; to be fitted. In many instances, the code will find that it has
;; pieces which cannot be fitted, and initiates a backtrack to remove
;; previously fitted pieces.
;; 
;; The standard code investigates 2,005 placements of the 18 pieces.
;; The speed of the standard code is highly dependent upon the
;; ordering of the pieces, which affects the ordering of the search; a
;; different ordering investigated 10 times as many placements, for
;; example. Interestingly enough, of the 2,005 partially-completed
;; puzzles investigated, 1,565 of them are distinct, meaning that
;; there is little hope of speedup from the "memoization" techniques
;; which have been found effective for other puzzles and games
;; [Bird80] [Baker92]. (The standard implementation of Puzzle
;; investigates surprisingly few configurations, making the ordering
;; of the puzzle pieces appear to have been tampered with to produce
;; shorter searches. See [Beeler84] for deeper analysis of the puzzle
;; itself.)
;; 
;; The standard Gabriel code for Puzzle does not have any errors, but
;; it does show evidence of a hasty conversion from a non-Lisp
;; language. It cannot decide, for example, whether to consistently
;; use 0-origin or 1-origin indexing. The standard code prefers to use
;; the more complex do instead of the simpler dotimes, and does not
;; utilize macros like incf and decf. None of these stylistic issues
;; should affect performance, however.
;; 
;; The one obvious stylistic change which could significantly improve
;; performance occurs at the end of the place routine where the puzzle
;; is searched for the smallest-numbered empty location. The Common
;; Lisp position "sequence" function could be used for this purpose,
;; and it could conceivably improve performance due to its presumably
;; high level of optimization.
;;
;; <URL:http://home.pipeline.com/~hbaker1/PuzzleB.html>


(declaim (optimize (debug 0)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defconstant +puzzle-size+ 511)
  (defconstant +puzzle-classmax+ 3)
  (defconstant +puzzle-dee+ 8)
  (defconstant +puzzle-typemax+ 12))

(defvar *iii* 0)
(defvar *kount* 0)

(declaim (type fixnum +puzzle-size+ +puzzle-classmax+ +puzzle-typemax+
               *iii* *kount* +puzzle-dee+))

(eval-when (:compile-toplevel :load-toplevel :execute)
  ;(deftype class-vector () `(simple-vector ,(1+ +puzzle-classmax+)))
  (deftype class-vector () `(simple-array fixnum (,(1+ +puzzle-classmax+))))
  ;(deftype type-vector () `(simple-vector ,(1+ +puzzle-typemax+)))
  (deftype type-vector () `(simple-array fixnum (,(1+ +puzzle-typemax+))))
  (deftype size-vector () `(simple-vector ,(1+ +puzzle-size+)))
  (deftype p-array () `(simple-array t (,(1+ +puzzle-typemax+) ,(1+ +puzzle-size+)))))

(defvar *piece-count* (make-array (1+ +puzzle-classmax+) :element-type 'fixnum :initial-element 0))
(defvar *class* (make-array (1+ +puzzle-typemax+) :element-type 'fixnum :initial-element 0))
(defvar *piecemax* (make-array (1+ +puzzle-typemax+) :element-type 'fixnum :initial-element 0))
(defvar *puzzle* (make-array (1+ +puzzle-size+)))
(defvar *puzzle-p* (make-array (list (1+ +puzzle-typemax+) (1+ +puzzle-size+))))

(declaim (type type-vector *class* *piecemax*))
(declaim (type class-vector *piece-count*))
(declaim (type size-vector *puzzle*))
(declaim (type p-array *puzzle-p*))

(defun fit (i j)
  (declare (fixnum i j))
  (let ((piecemax *piecemax*)
	(puzzle-p *puzzle-p*)
	(puzzle *puzzle*)
	)
    (declare (type type-vector piecemax) (type p-array puzzle-p) (type size-vector puzzle))
    (dotimes (k (#|svref|# aref piecemax i))
      (declare (fixnum k))
      (and (aref puzzle-p i k)
	   (#|svref|# aref puzzle (the fixnum (+ j k)))
	   (return nil))))
  t)

(defun place (i j &aux (piecemax *piecemax*))
  (declare (fixnum i j) (type type-vector piecemax))
  (let ((end (#|svref|# aref piecemax i))
	(piece-count *piece-count*)
	(class *class*)
	(puzzle-p *puzzle-p*)
	(puzzle *puzzle*)
	)
    (declare (fixnum end)
	     (type class-vector piece-count) (type type-vector class)
	     (type p-array puzzle-p) (type size-vector puzzle))
    (dotimes (k end)
      (declare (fixnum k))
      (when (aref puzzle-p i k)
        (setf (#|svref|# aref puzzle (the fixnum (+ j k))) t)))
    (setf (#|svref|# aref piece-count (the fixnum (#|svref|# aref class i)))
	  (the fixnum (- (the fixnum (#|svref|# aref piece-count (#|svref|# aref class i))) 1)))
    (do ((k j (1+ k)))
	((> k +puzzle-size+)
;	 (terpri)
;	 (princ "Puzzle filled")
	 0)
	(declare (fixnum k))
      (unless (#|svref|# aref puzzle k)
        (return k)))))

(defun puzzle-remove (i j &aux (piecemax *piecemax*))
  (declare (fixnum i j) (type type-vector piecemax))
  (let ((end (#|svref|# aref piecemax i))
	(piece-count *piece-count*)
	(class *class*)
	(puzzle-p *puzzle-p*)
	(puzzle *puzzle*)
	)
    (declare (fixnum end)
	     (type class-vector piece-count) (type type-vector class)
	     (type p-array puzzle-p) (type size-vector puzzle))
    (do ((k 0 (1+ k)))
	((> k end))
	(declare (fixnum k))
      (when (aref puzzle-p i k)
        (setf (#|svref|# aref puzzle (the fixnum (+ j k)))  nil))
      (setf (#|svref|# aref piece-count (the fixnum (#|svref|# aref class i)))
	    (the fixnum (+ (the fixnum (#|svref|# aref piece-count (#|svref|# aref class i)))
			   1))))))

(defun trial (j)
  (declare (fixnum j))
  (let ((k 0) (piece-count *piece-count*) (class *class*))
    (declare (fixnum k) (type class-vector piece-count) (type type-vector class))
    (do ((i 0 (1+ i)))
	((> i +puzzle-typemax+)
         (incf *kount*)
	 nil)
	(declare (fixnum i))
      (unless (eql (#|svref|# aref piece-count (#|svref|# aref class i)) 0)
        (when (fit i j)
          (setq k (place i j))
          (cond ((or (trial k)
                     (= k 0))
                 ;; (format t "~%Piece ~4D at ~4D." (+ i 1) (+ k 1))
                 (incf *kount*)
                 (return t))
                (t (puzzle-remove i j))))))))

(defun definepiece (iclass ii jj kk)
  (declare (fixnum ii jj kk))
  (let ((index 0) (iii *iii*) (puzzle-p *puzzle-p*))
    (declare (fixnum index iii) (type p-array puzzle-p))
    (do ((i 0 (1+ i)))
	((> i ii))
	(declare (fixnum i))
      (do ((j 0 (1+ j)))
	  ((> j jj))
	  (declare (fixnum j))
	(do ((k 0 (1+ k)))
	    ((> k kk))
	    (declare (fixnum k))
	  (setq index  (+ i (the fixnum
				 (* +puzzle-dee+ (the fixnum
						      (+ j (the fixnum (* +puzzle-dee+ k))))))))
	  (setf (aref puzzle-p iii index)  t))))
    (let ((class *class*))
      (declare (type type-vector class))
      (setf (#|svref|# aref class iii) iclass))
    (let ((piecemax *piecemax*))
      (declare (type type-vector piecemax))
      (setf (#|svref|# aref piecemax iii) index))
    (unless (= iii +puzzle-typemax+)
      ;(incf *iii*)
      (setq *iii* (setq iii (1+ iii)))
      )))

(defun run-puzzle (&aux (puzzle *puzzle*))
  (declare (type size-vector puzzle))
  (do ((m 0 (1+ m)))
      ((> m +puzzle-size+))
      (declare (fixnum m))
    (setf (#|svref|# aref puzzle m) t))
  (do ((i 1 (1+ i)))
      ((> i 5))
      (declare (fixnum i))
    (do ((j 1 (1+ j)))
	((> j 5))
	(declare (fixnum j))
      (do ((k 1 (1+ k)))
	  ((> k 5))
	  (declare (fixnum k))
	(setf (#|svref|# aref puzzle
		     (the fixnum (+ i
				    (the fixnum
					 (* +puzzle-dee+ (the fixnum
							      (+ j (the fixnum (* +puzzle-dee+ k)))))))))
	      nil))))
  (do ((i 0 (1+ i))
       (puzzle-p *puzzle-p*)
       )
      ((> i +puzzle-typemax+))
      (declare (type p-array puzzle-p) (fixnum i))
    (do ((m 0 (1+ m)))
	((> m +puzzle-size+))
	(declare (fixnum m))
      (setf (aref puzzle-p i m)  nil)))
  (setq *iii* 0)
  (definePiece 0 3 1 0)
  (definePiece 0 1 0 3)
  (definePiece 0 0 3 1)
  (definePiece 0 1 3 0)
  (definePiece 0 3 0 1)
  (definePiece 0 0 1 3)

  (definePiece 1 2 0 0)
  (definePiece 1 0 2 0)
  (definePiece 1 0 0 2)

  (definePiece 2 1 1 0)
  (definePiece 2 1 0 1)
  (definePiece 2 0 1 1)

  (definePiece 3 1 1 1)

  (let ((piece-count *piece-count*))
    (declare (type class-vector piece-count))
    (setf (#|svref|# aref piece-count 0) 13.)
    (setf (#|svref|# aref piece-count 1) 3)
    (setf (#|svref|# aref piece-count 2) 1)
    (setf (#|svref|# aref piece-count 3) 1))
  (let ((m (1+ (* +puzzle-dee+ (1+ +puzzle-dee+))))
	(n 0)
	(*kount* 0))
    (declare (fixnum m n))
    (cond ((fit 0 m) (setq n (place 0 m)))
	  (t (format t "~%Error.")))
    (cond ((trial n)
	   ;; (format t "~%Success in ~4D trials." kount)
           )
	  (t (format t "~%Failure.")))))




(defun full-gc ()
  #+mkcl (si:gc)
  #+ecl (si:gc t)
  #+sbcl (sb-ext:gc #+gencgc :full #+gencgc t)
  #+clisp (gc)
  )

(defun run-test ()
  (format t "~&Running run-puzzle.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1500) (declare (fixnum i)) (run-puzzle)))
  )
